﻿Imports System.Collections
''' <summary>
''' A collection of sections of a ini file.
''' </summary>
Public Class IniFileSectionCollection
    Inherits ObjectModel.Collection(Of IniFileSection)

    Private iecHeaderComparer As IEqualityComparer(Of String)
    Private ifIniFile As IniFile

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSectionCollection class.
    ''' </summary>
    ''' <param name="iniFile">The ini file whose sections are stored in this instance of the IniLib.IniFileSectionCollection class.</param>
    ''' <param name="sectionNameComparer">The comparer which to use when comparing two section names.</param>
    Public Sub New(ByVal iniFile As IniFile, ByVal sectionNameComparer As IEqualityComparer(Of String))
        iecHeaderComparer = sectionNameComparer
        ifIniFile = iniFile
    End Sub

    ''' <summary>
    ''' Inserts a section at the specified index.
    ''' </summary>
    ''' <param name="index">The index at which to insert the section.</param>
    ''' <param name="item">The section to insert at the specified index.</param>
    Protected Overrides Sub InsertItem(ByVal index As Integer, ByVal item As IniFileSection)
        If item.IsSectionNameVisible AndAlso (ifIniFile.Flags And IniFileFlags.SupportsSectionNameRepetition) = IniFileFlags.Default AndAlso Me.Contains(item.SectionName) Then
            Throw New ArgumentException("Collection already contains an item with the specified section name.")
        End If
        MyBase.InsertItem(index, item)
    End Sub

    ''' <summary>
    ''' Replaces the section currently at the specified index by the specified item.
    ''' </summary>
    ''' <param name="index">The index at which to set the section.</param>
    ''' <param name="item">The section replacing the section which currently is at the index specified by index.</param>
    Protected Overrides Sub SetItem(ByVal index As Integer, ByVal item As IniFileSection)
        If item.IsSectionNameVisible AndAlso (IniFile.Flags And IniFileFlags.SupportsSectionNameRepetition) = IniFileFlags.Default AndAlso Me.Contains(item.SectionName) AndAlso Not iecHeaderComparer.Equals(item.SectionName, Me.Items(index).SectionName) Then
            Throw New ArgumentException("Collection already contains an item with the specified key.")
        End If
        MyBase.SetItem(index, item)
    End Sub

    ''' <summary>
    ''' The parent file of all sections stored in this instance of IniLib.IniFileSectionCollection class.
    ''' </summary>
    Public ReadOnly Property IniFile As IniFile
        Get
            Return ifIniFile
        End Get
    End Property


    ''' <summary>
    ''' Retrieves whether the section collection contains a section with the specified section name.
    ''' </summary>
    ''' <param name="sectionName">The section name to search in the collection.</param>rns>
    ''' <remarks></remarks>
    Public Overloads Function Contains(ByVal sectionName As String) As Boolean
        For Each c As IniFileSection In Me
            If iecHeaderComparer.Equals(c.SectionName, sectionName) Then
                Return True
            End If
        Next
        Return False
    End Function

    ''' <summary>
    ''' Gets the index of the first section whose section name equals the specified section name.
    ''' </summary>
    ''' <param name="sectionName">The section name whose index to get.</param>
    ''' <returns>Returns either the index of the section name with the specified key or -1 if there aren't any sections with the specified section name.</returns>
    Public Overloads Function IndexOf(ByVal sectionName As String) As Integer
        For i As Integer = 0 To Count - 1
            If iecHeaderComparer.Equals(Me.Item(i).SectionName, sectionName) Then
                Return i
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Gets the index of the first section whose section name equals the specified section name starting at the specified start index.
    ''' </summary>
    ''' <param name="sectionName">The section name to look for.</param>
    ''' <param name="startIndex">The index at which to start searching for a section with the specified section name.</param>
    ''' <returns>Returns either the index of the section with the specified section name or -1 if there aren't any sections with the specified section name.</returns>
    Public Overloads Function IndexOf(ByVal sectionName As String, ByVal startIndex As Integer) As Integer
        For i As Integer = startIndex To Count - 1
            If iecHeaderComparer.Equals(Me.Item(i).SectionName, sectionName) Then
                Return i
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Gets the first section with the specified section name.
    ''' </summary>
    ''' <param name="sectionName">The section name of the section to look for.</param>
    ''' <value></value>
    ''' <returns>Returns either the first section with the specified section name or null if there aren't any sections with the specified section name.</returns>
    Default Public Overloads ReadOnly Property Item(ByVal sectionName As String) As IniFileSection
        Get
            For i As Integer = 0 To Count - 1
                If iecHeaderComparer.Equals(Me.Item(i).SectionName, sectionName) Then
                    Return Item(i)
                End If
            Next
            Return Nothing
        End Get
    End Property

    ''' <summary>
    ''' Invokes the specified action for all sections in the collection whose section name is equal to the specified section name.
    ''' </summary>
    ''' <param name="sectionName">The section name of the sections to invoke the specified action for.</param>
    ''' <param name="action">The action to invoke for all sections with the specified section name in the collection.</param>
    ''' <remarks></remarks>
    Public Sub ForEach(ByVal sectionName As String, ByVal action As Action(Of IniFileSection))
        For Each itm As IniFileSection In Me
            If iecHeaderComparer.Equals(itm.SectionName, sectionName) Then
                action(itm)
            End If
        Next
    End Sub

    ''' <summary>
    ''' Returns the comparer to use when comparing two sections' names.
    ''' </summary>
    Public ReadOnly Property Comparer As IEqualityComparer(Of String)
        Get
            Return iecHeaderComparer
        End Get
    End Property
End Class
