﻿Imports System.Text

''' <summary>
''' Provides the base class for reading ini files.
''' </summary>
''' <remarks></remarks>
Public MustInherit Class IniTextReader
    Inherits TextReader

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniTextReader class.
    ''' </summary>
    Public Sub New()
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniTextReader class setting the amount of chars to be stored in a string builder used during general operations on reading strings.
    ''' </summary>
    Public Sub New(ByVal stringBufferDefaultLength As Integer)
        MyBase.New(stringBufferDefaultLength)
    End Sub

    ''' <summary>
    ''' Returns the character separating key and value. This property may be overriden when it comes to files that are structured similarly to ini files.
    ''' </summary>
    Public Overridable ReadOnly Property KeyValueSeparator As Char
        Get
            Return "="c
        End Get
    End Property

    ''' <summary>
    ''' Returns the character separating key-value pairs and comments. This property may be overriden when it comes to files that are structured similarly to ini files.
    ''' </summary>
    Public Overridable ReadOnly Property CommentSeparator As Char
        Get
            Return ";"c
        End Get
    End Property

    ''' <summary>
    ''' Retrieves whether values are allowed to be empty.
    ''' </summary>
    Protected Overridable ReadOnly Property AllowEmptyValues As Boolean
        Get
            Return False
        End Get
    End Property

    ''' <summary>
    ''' Reads the key of a key-value pair.
    ''' </summary>
    Public Function ReadKey() As String
        SkipSpaces()
        Dim chr As Char = CurrentChar()
        Dim sb As New StringBuilder(StringBufferDefaultLength)

        Dim trailingWhitespaceCount As Integer

        Do
            If IsValidContentChar(chr) Then
                sb.Append(chr)
            ElseIf IsLinefeedChar(chr) Then
                Throw New IniFileSyntaxException("Line feed characters are not valid in keys.")
            Else
                Throw New IniFileSyntaxException("Character '" & chr & "' is not a valid character.")
            End If
            If Char.IsWhiteSpace(chr) Then
                trailingWhitespaceCount += 1
            Else
                trailingWhitespaceCount = 0
            End If
            chr = ReadChar()
        Loop Until chr = KeyValueSeparator

        sb.Remove(sb.Length - trailingWhitespaceCount, trailingWhitespaceCount)
        If sb.Length <> 0 Then
            Return sb.ToString
        Else
            Throw New IniFileSyntaxException("Key mustnot be empty.")
        End If
    End Function

    ''' <summary>
    ''' Reads the value of a key-value pair.
    ''' </summary>
    Public Function ReadValue() As String
        Dim chr As Char = CurrentChar()
        Dim sb As New StringBuilder(StringBufferDefaultLength)

        Dim trailingWhitespaceCount As Integer
        If chr = KeyValueSeparator Then
            MoveNext()
            SkipSpaces()
            chr = CurrentChar
        Else
            Throw New IniFileSyntaxException("Value must be separated from key by a '" & KeyValueSeparator & "'-character.")
        End If
        If (IsLinefeedChar(chr) OrElse chr = CommentSeparator) Then
            If AllowEmptyValues() Then
                Return String.Empty
            Else
                Throw New IniFileSyntaxException("Empty values are not allowed.")
            End If
        Else
            Do
                If IsValidContentChar(chr) Then
                    sb.Append(chr)
                Else
                    Throw New IniFileSyntaxException("Character '" & chr & "' is not a valid character.")
                End If
                If Not chr = " "c OrElse chr = Microsoft.VisualBasic.Chr(9) Then
                    trailingWhitespaceCount = 0
                Else
                    trailingWhitespaceCount += 1
                End If
                If Not EndOfContent Then
                    chr = ReadChar()
                Else
                    Exit Do
                End If
            Loop Until (IsLinefeedChar(chr) OrElse chr = CommentSeparator)

            sb.Remove(sb.Length - trailingWhitespaceCount, trailingWhitespaceCount)
            If sb.Length <> 0 Then
                Return sb.ToString
            Else
                Throw New IniFileSyntaxException("Empty values are not allowed.")
            End If
        End If
    End Function

    ''' <summary>
    ''' Reads comments in ini files.
    ''' </summary>
    Public Function ReadComment() As String
        SkipSpaces()
        Dim chr As Char = CurrentChar
        Dim sb As New StringBuilder(StringBufferDefaultLength)

        If chr = CommentSeparator Then
            chr = ReadChar()
            If IsLinefeedChar(chr) Then
                Return String.Empty
            End If
        Else
            Throw New IniFileSyntaxException("Value must be separated from key by a '" & CommentSeparator & "'-character.")
        End If
        Do
            If IsValidCommentChar(chr) Then
                sb.Append(chr)
            Else
                Throw New IniFileSyntaxException("Character '" & chr & "' is not a valid character.")
            End If
            If Not EndOfContent Then
                chr = ReadChar()
            Else
                Exit Do
            End If
        Loop Until IsLinefeedChar(chr)

        Return sb.ToString()
    End Function

    ''' <summary>
    ''' Reads a key-value pair and a possible comment as a IniLib.IniKeyValuePairDescription.
    ''' </summary>
    Public Function ReadIniKeyValuePair() As IniKeyValuePairDescription
        Dim ikvpd As New IniKeyValuePairDescription(ReadKey(), ReadValue())
        If CurrentChar = CommentSeparator Then
            ikvpd.Comment = ReadComment()
        End If
        Return ikvpd
    End Function

    ''' <summary>
    ''' Reads the section name of an ini section.
    ''' </summary>
    Public Function ReadIniSectionHeader() As String
        SkipSpaces()
        Dim chr As Char = CurrentChar()
        Dim sb As New StringBuilder(StringBufferDefaultLength)
        Dim ws As Boolean
        If chr <> "["c Then
            Throw New IniFileSyntaxException("Ini section must start with '['.")
        End If

        chr = ReadChar()
        If chr = "]"c Then
            Throw New IniFileSyntaxException("Unexpected end of ini section name. Section name is not allowed to be empty.")
        ElseIf Char.IsWhiteSpace(chr) Then
            Throw New IniFileSyntaxException("Section name is not allowed to start with a whitespace character.")
        End If
        Do
            If IsValidContentChar(chr) Then
                sb.Append(chr)
                ws = Char.IsWhiteSpace(chr)
            ElseIf IsLinefeedChar(chr) Then
                Throw New IniFileSyntaxException("Line feed characters are not valid in section names.")
            Else
                Throw New IniFileSyntaxException("Character '" & chr & "' is not a valid character.")
            End If
            chr = ReadChar()
        Loop Until chr = "]"c
        MoveNext()
        If ws Then
            Throw New IniFileSyntaxException("Trailing whitespaces in section name are not allowed.")
        End If

        Return sb.ToString()
    End Function

    ''' <summary>
    ''' Determines whether the specified character is valid in the content of an ini file.
    ''' </summary>
    ''' <param name="charValue">Specifies the character to check.</param>
    Protected Overridable Function IsValidContentChar(ByVal charValue As Char) As Boolean
        Return (Not (IsLinefeedChar(charValue)))
    End Function
    
    ''' <summary>
    ''' Determines whether the specified character is valid in comments of an ini file.
    ''' </summary>
    ''' <param name="charValue">Specifies the character to check.</param>
    Protected Overridable Function IsValidCommentChar(ByVal charValue As Char) As Boolean
        Return (Not IsLinefeedChar(charValue))
    End Function
End Class
