﻿Imports System.Text, Microsoft.VisualBasic

''' <summary>
''' Provides a base class for reading texts from different sources.
''' </summary>
Public MustInherit Class TextReader

    Private intStringBufferDefaultLength As Integer

    ''' <summary>
    ''' Moves on to the next character.
    ''' </summary>
    Public MustOverride Sub MoveNext()

    ''' <summary>
    ''' Reads characters into the specified buffer starting at the specified index. Returns the amount of characters read.
    ''' </summary>
    ''' <param name="chars">The buffer to fill with characters.</param>
    ''' <param name="index">The index at which to start filling the buffer.</param>
    ''' <param name="count">The maximum amount of characters to read.</param>
    ''' <returns>Returns the amount of characters read into the buffer.</returns>
    Public MustOverride Overloads Function ReadChars(ByVal chars() As Char, ByVal index As Integer, ByVal count As Integer) As Integer

    ''' <summary>
    ''' Reads all characters in the stream starting from the current offset.
    ''' </summary>
    Public MustOverride Function ReadToEnd() As String

    ''' <summary>
    ''' Retrieves whether the end of content is reached.
    ''' </summary>
    Public MustOverride ReadOnly Property EndOfContent() As Boolean

    ''' <summary>
    ''' Returns the character currently selected.
    ''' </summary>
    Public MustOverride ReadOnly Property CurrentChar() As Char

    ''' <summary>
    ''' Reads a char-array containing the specified amount of characters.
    ''' </summary>
    ''' <param name="count"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overridable Overloads Function ReadChars(ByVal count As Integer) As Char()
        Dim c(count - 1) As Char
        ReadChars(c, 0, count)
        Return c
    End Function

    ''' <summary>
    ''' Fills the specified array with characters and returns the amount of read characters.
    ''' </summary>
    ''' <param name="buffer">The buffer to fill with characters.</param>
    Public Overridable Overloads Function ReadChars(ByVal buffer() As Char) As Integer
        Return ReadChars(buffer, 0, buffer.Length)
    End Function

    ''' <summary>
    ''' Reads a string with the specified length.
    ''' </summary>
    ''' <param name="length">The amount of characters to read.</param>
    Public Overridable Function ReadString(ByVal length As Integer) As String
        Dim sb As New StringBuilder(length)
        Dim chr As Char
        While (sb.Length < length)
            If TryReadChar(chr) Then
                sb.Append(chr)
            Else
                Return sb.ToString
            End If

        End While

        Return sb.ToString()
    End Function

    ''' <summary>
    ''' Specifies whether the text reader currently is able to read.
    ''' </summary>
    Protected Overridable ReadOnly Property CanRead() As Boolean
        Get
            Return Not EndOfContent()
        End Get
    End Property

    ''' <summary>
    ''' Tries to read a character which, if successful, is stored in the result parameter and returns whether the operation was successful.
    ''' </summary>
    ''' <param name="result">Contains the read character if the operation was successful. Otherwise the parameter's value does not change.</param>
    Public Overridable Function TryReadChar(ByRef result As Char) As Boolean
        If EndOfContent() = False Then
            result = ReadChar()
            Return True
        Else
            Return False
        End If
    End Function

    ''' <summary>
    ''' Returns the default length of a string builder which is instantiated during general string reading operations.
    ''' </summary>
    Public ReadOnly Property StringBufferDefaultLength() As Integer
        Get
            Return intStringBufferDefaultLength
        End Get
    End Property

    ''' <summary>
    ''' Initializes a new instance of the IniLib.TextReader class.
    ''' </summary>
    ''' <param name="stringBufferDefaultLength">Specifies the default length of a string builder which is instantiated during general string reading operations.</param>
    Public Sub New(ByVal stringBufferDefaultLength As Integer)
        intStringBufferDefaultLength = stringBufferDefaultLength
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.TextReader class.
    ''' </summary>
    Public Sub New()
        intStringBufferDefaultLength = 8192
    End Sub

    ''' <summary>
    ''' Reads a line.
    ''' </summary>
    Public Overridable Function ReadLine() As String
        ReadLine = ReadUntil(AddressOf IsLinefeedChar)
        Advance()
    End Function
    
    ''' <summary>
    ''' Skips all characters while they match the specified requirements.
    ''' </summary>
    ''' <param name="match">A delegate specifying the requirements.</param>
    Public Overridable Sub SkipWhile(ByVal match As Predicate(Of Char))
        If match(CurrentChar) Then
            While Not EndOfContent() AndAlso match(CurrentChar)
                MoveNext()
            End While
        End If
    End Sub

    ''' <summary>
    ''' Skips line feeds.
    ''' </summary>
    ''' <remarks>As there are differences in treating line feed characters when it comes to different standards the IniLib supports these. The IniLib's rules of interpreting line feeds are as follows.
    ''' Cr, Lf, CrLf, LfCr are interpreted as one character (Cr = Carriage return, Lf = Line feed) while LfLf or CrCr are two characters.</remarks>
    Public Sub SkipLinefeed()
        If CurrentChar = Chr(10) Then
            MoveNext()
            If CurrentChar = Chr(13) Then
                MoveNext()
            End If
        ElseIf CurrentChar = Chr(13) Then
            MoveNext()
            If CurrentChar = Chr(10) Then
                MoveNext()
            End If
        Else
            Throw New ArgumentException("Current char was not a linefeed character.")
        End If
    End Sub

    ''' <summary>
    ''' Reads a string while the match parameter indicates that a character matches the requirements.
    ''' </summary>
    ''' <param name="match">A delegate specifying the requirements of the characters.</param>
    Public Overridable Function ReadWhile(ByVal match As Predicate(Of Char)) As String
        If match(CurrentChar) Then
            Dim sb As New StringBuilder(StringBufferDefaultLength)
            While Not EndOfContent() AndAlso match(CurrentChar)
                sb.Append(CurrentChar)
                MoveNext()
            End While
            Return sb.ToString
        Else
            Return CurrentChar.ToString
        End If
    End Function

    ''' <summary>
    ''' Reads a string until the match parameter indicates that a character does not match the requirements.
    ''' </summary>
    ''' <param name="match">A delegate specifying the requirements of the characters.</param>
    Public Overridable Function ReadUntil(ByVal match As Predicate(Of Char)) As String
        If Not match(CurrentChar) Then
            Dim sb As New StringBuilder(StringBufferDefaultLength)
            While Not (EndOfContent() OrElse match(CurrentChar))
                sb.Append(CurrentChar)
                MoveNext()
            End While
            Return sb.ToString
        Else
            Return CurrentChar.ToString
        End If
    End Function

    ''' <summary>
    ''' Reads a string until the reached character matches the specified character.
    ''' </summary>
    ''' <param name="character">The character which represents the end of the string to retrieve.</param>
    Public Overridable Function ReadUntil(ByVal character As Char) As String
        If CurrentChar <> character Then
            Dim sb As New StringBuilder(StringBufferDefaultLength)
            While Not (EndOfContent() OrElse CurrentChar = character)
                sb.Append(CurrentChar)
                MoveNext()
            End While
            Return sb.ToString
        Else
            Return CurrentChar.ToString
        End If
    End Function

    ''' <summary>
    ''' Advances the buffer position by one and returns the character which is then at the buffer position.
    ''' </summary>
    Public Overridable Function ReadChar() As Char
        MoveNext()
        Return CurrentChar
    End Function

    ''' <summary>
    ''' Reads a word.
    ''' </summary>
    Public Overridable Function ReadWord() As String
        Return ReadWhile(Function(charValue As Char) Char.IsWhiteSpace(charValue) OrElse Char.IsPunctuation(charValue) OrElse Char.IsSeparator(charValue))
    End Function

    ''' <summary>
    ''' Sets the position of the underlying stream to the actual position considering the buffer position.
    ''' </summary>
    Public Overridable Sub SkipSpaces()
        If CurrentChar = " "c OrElse CurrentChar = Chr(9) Then
            While (CurrentChar = Chr(9) OrElse CurrentChar = " "c)
                MoveNext()
            End While
        End If
    End Sub

    ''' <summary>
    ''' Advances the buffer position depending on the content. View remarks for further information.
    ''' </summary>
    ''' <returns>Returns whether advancing was successful.</returns>
    ''' <remarks>As there are differences in treating line feed characters when it comes to different standards the IniLib supports these. The IniLib's rules of interpreting line feeds are as follows.
    ''' Cr, Lf, CrLf, LfCr are interpreted as one character (Cr = Carriage return, Lf = Line feed) while LfLf or CrCr are two characters.</remarks>
    Public Overridable Function Advance() As Boolean
        If Not EndOfContent Then
            If IsLinefeedChar(CurrentChar) Then
                SkipLinefeed()
            Else
                MoveNext()
            End If
            Return Not EndOfContent
        Else
            Return False
        End If
    End Function

    ''' <summary>
    ''' Flushes this text reader.
    ''' </summary>
    Public Overridable Sub Flush()
    End Sub

    ''' <summary>
    ''' Indicates whether the specified character is a line feed character.
    ''' </summary>
    ''' <param name="charValue">The character to check.</param>
    Protected Overridable Function IsLinefeedChar(ByVal charValue As Char) As Boolean
        Return charValue = Chr(10) OrElse charValue = Chr(13)
    End Function
End Class
