﻿Imports System.Collections
Imports System.Text, Microsoft.VisualBasic

''' <summary>
''' Manages the structure of ini files.
''' </summary>
Public Class IniFile
    Private iffFlags As IniFileFlags
    Private ifscSections As IniFileSectionCollection

    ''' <summary>
    ''' Gets the flags passed to the constructor of this instance of IniLib.IniFile. View <seealso cref="IniLib.IniFileFlags">IniLib.IniFileFlags</seealso> for further information.
    ''' </summary>
    Public ReadOnly Property Flags As IniFileFlags
        Get
            Return iffFlags
        End Get
    End Property

    ''' <summary>
    ''' Gets a collection of ini file sections that describe the content of the ini file.
    ''' </summary>
    Public ReadOnly Property Sections As IniFileSectionCollection
        Get
            Return ifscSections
        End Get
    End Property

    ''' <summary>
    ''' Creates the section collection of the ini file. Override this method to customize ini section management.
    ''' </summary>
    Protected Friend Overridable Function CreateSectionCollection() As IniFileSectionCollection
        If (iffFlags And IniFileFlags.CaseInsensitive) = IniFileFlags.CaseInsensitive Then
            Return New IniFileSectionCollection(Me, StringComparer.InvariantCultureIgnoreCase)
        Else
            Return New IniFileSectionCollection(Me, StringComparer.InvariantCulture)
        End If
    End Function

    ''' <summary>
    ''' Creates key-value pair collections of ini sections. Override this method to customize ini key-value pair management.
    ''' </summary>
    ''' <param name="section">Specifies the section that receives the collection created in the method.</param>
    Protected Friend Overridable Function CreateKeyValuePairCollection(ByVal section As IniFileSection) As IniFileKeyValuePairCollection
        If (iffFlags And IniFileFlags.CaseInsensitive) = IniFileFlags.CaseInsensitive Then
            Return New IniFileKeyValuePairCollection(section, StringComparer.InvariantCultureIgnoreCase)
        Else
            Return New IniFileKeyValuePairCollection(section, StringComparer.InvariantCulture)
        End If
    End Function

    ''' <summary>
    ''' Loads an ini-file keeping the current content.
    ''' </summary>
    ''' <param name="fileName">The file name of the ini-file to be loaded.</param>
    Public Sub LoadFile(ByVal fileName As String)
        LoadFile(fileName, True)
    End Sub

    ''' <summary>
    ''' Loads an ini-file.
    ''' </summary>
    ''' <param name="fileName">The file name of the ini-file to be loaded.</param>
    ''' <param name="keepPrevious">Specifies whether to clear the content before adding new keys.</param>
    Public Sub LoadFile(ByVal fileName As String, ByVal keepPrevious As Boolean)
        Dim fs As IO.FileStream = New IO.FileStream(fileName, IO.FileMode.Open)
        LoadStream(fs, keepPrevious)
        fs.Close()
    End Sub

    ''' <summary>
    ''' Loads an ini-file keeping the current content.
    ''' </summary>
    ''' <param name="fileName">The file name of the ini-file to be loaded.</param>
    Public Sub LoadFile(ByVal fileName As String, ByVal encoding As Encoding)
        LoadFile(fileName, encoding, True)
    End Sub

    ''' <summary>
    ''' Loads an ini-file.
    ''' </summary>
    ''' <param name="fileName">The file name of the ini-file to be loaded.</param>
    ''' <param name="keepPrevious">Specifies whether to clear the content before adding new keys.</param>
    Public Sub LoadFile(ByVal fileName As String, ByVal encoding As Encoding, ByVal keepPrevious As Boolean)
        Dim fs As IO.FileStream = New IO.FileStream(fileName, IO.FileMode.Open)
        LoadStream(fs, encoding, keepPrevious)
        fs.Close()
    End Sub

    ''' <summary>
    ''' Returns the character used to separate keys and values.
    ''' </summary>
    Public Overridable ReadOnly Property KeyValueSeparator As Char
        Get
            Return "="c
        End Get
    End Property

    ''' <summary>
    ''' Returns the character used to separate key-value pairs and comments.
    ''' </summary>
    Public Overridable ReadOnly Property CommentSeparator As Char
        Get
            Return ";"c
        End Get
    End Property

    ''' <summary>
    ''' Returns whether the specified section name is valid.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name to check.</param>
    Public Overridable Function IsValidSectionName(ByVal sectionName As String) As Boolean
        If Not String.IsNullOrEmpty(sectionName) Then
            Dim ws As Boolean
            If Char.IsWhiteSpace(sectionName(0)) Then
                Return False
            End If

            For Each c As Char In sectionName
                If c = Chr(13) OrElse c = Chr(10) OrElse c = "]"c OrElse c = "["c OrElse c = KeyValueSeparator OrElse c = CommentSeparator Then
                    Return False
                End If
                ws = Char.IsWhiteSpace(c)
            Next
            Return Not ws
        Else
            Return True
        End If
    End Function

    ''' <summary>
    ''' Returns whether the specified key is valid.
    ''' </summary>
    ''' <param name="key">Specifies the key to check.</param>
    Public Overridable Function IsValidKeyName(ByVal key As String) As Boolean
        If key.Length <> 0 Then
            Dim ws As Boolean
            If Char.IsWhiteSpace(key(0)) Then
                Return False
            End If
            For Each c As Char In key
                If c = Chr(13) OrElse c = Chr(10) OrElse c = "]"c OrElse c = "["c OrElse c = KeyValueSeparator OrElse c = CommentSeparator Then
                    Return False
                End If
                ws = Char.IsWhiteSpace(c)
            Next
            Return Not ws
        Else
            Return False
        End If
    End Function

    ''' <summary>
    ''' Returns whether the specified comment is valid.
    ''' </summary>
    ''' <param name="comment">Specifies the comment to check.</param>
    Public Overridable Function IsValidComment(ByVal comment As String) As Boolean
        If comment.Length <> 0 Then
            For Each c As Char In comment
                If c = Chr(13) OrElse c = Chr(10) Then
                    Return False
                End If
            Next
            Return True
        Else
            Return True
        End If
    End Function

    ''' <summary>
    ''' Loads the ini data from the specified stream and keeps the previously loaded data.
    ''' </summary>
    ''' <param name="stream">Specifies the stream from which to load the ini data.</param>
    Public Sub LoadStream(ByVal stream As IO.Stream)
        LoadStream(stream, True)
    End Sub

    ''' <summary>
    ''' Loads the ini data from the specified stream and clears the previously loaded data if specified.
    ''' </summary>
    ''' <param name="stream">Specifies the stream from which to load the ini data.</param>
    ''' <param name="keepPrevious">Specifies whether to keep previously loaded data.</param>
    Public Sub LoadStream(ByVal stream As IO.Stream, ByVal keepPrevious As Boolean)
        SyncLock stream
            Load(New IniFileCustomReader(stream, Encoding.Default, CommentSeparator, KeyValueSeparator), keepPrevious)
        End SyncLock
    End Sub

    ''' <summary>
    ''' Loads the ini data from the specified stream that is encoded as specified and keeps previously loaded ini data.
    ''' </summary>
    ''' <param name="stream">Specifies the stream from which to load ini data.</param>
    ''' <param name="encoding">Specifies the encoding of the data stored in the stream.</param>
    Public Sub LoadStream(ByVal stream As IO.Stream, ByVal encoding As Encoding)
        LoadStream(stream, encoding, True)
    End Sub

    ''' <summary>
    ''' Loads the ini data from the specified stream that is encoded as specified and keeps previously loaded ini data.
    ''' </summary>
    ''' <param name="stream">Specifies the stream from which to load ini data.</param>
    ''' <param name="encoding">Specifies the encoding of the data stored in the stream.</param>
    ''' <param name="keepPrevious">Specifies whether to keep previously loaded ini data.</param>
    Public Sub LoadStream(ByVal stream As IO.Stream, ByVal encoding As Encoding, ByVal keepPrevious As Boolean)
        SyncLock stream
            Dim streamReader As IniTextReader = New IniFileCustomReader(stream, encoding, CommentSeparator, KeyValueSeparator)
            Load(streamReader, keepPrevious)
            streamReader.Flush()
        End SyncLock
    End Sub

    ''' <summary>
    ''' Loads ini data from the specified ini text reader and keeps previously read ini data.
    ''' </summary>
    ''' <param name="reader">Specifies the reader from which to read ini data.</param>
    Public Sub Load(ByVal reader As IniTextReader)
        Load(reader, True)
    End Sub

    ''' <summary>
    ''' Loads ini data from the specified ini text reader and clears previously read ini data if specified.
    ''' </summary>
    ''' <param name="reader">Specifies the reader from which to read ini data.</param>
    ''' <param name="keepPrevious">Specifies whether to clear previously loaded ini data.</param>
    Public Sub Load(ByVal reader As IniTextReader, ByVal keepPrevious As Boolean)
        If keepPrevious = False Then
            Sections.Clear()
        End If
        Dim count As Integer = Sections.Count
        Dim cs As String = Nothing
        Dim ifsc As IniFileSection = Nothing
        Dim kvp As IniFileKeyValuePair
        If Not reader.EndOfContent Then
            Do
                reader.SkipSpaces()
                If reader.CurrentChar = "["c Then
                    If cs IsNot Nothing Then
                        ifsc.KeyValues.Add(New IniFileKeyValuePair(ifsc, Nothing, Nothing, cs))
                    End If
                    ifsc = New IniFileSection(Me, reader.ReadIniSectionHeader())
                    If reader.CurrentChar = ";"c Then
                        ifsc.Comment = reader.ReadComment()
                    End If
                    cs = Nothing
                    Me.Sections.Add(ifsc)
                ElseIf reader.CurrentChar = ";"c Then
                    cs = reader.ReadComment()
                    If ifsc Is Nothing Then
                        ifsc = New IniFileSection(Me, Nothing, cs)
                        Me.Sections.Add(ifsc)
                        cs = Nothing
                    End If
                ElseIf reader.CurrentChar = Chr(13) OrElse reader.CurrentChar = Chr(10) Then
                    If ifsc Is Nothing Then
                        ifsc = New IniFileSection(Me, Nothing, String.Empty)
                        Me.Sections.Add(ifsc)
                    Else
                        ifsc.KeyValues.Add(New IniFileKeyValuePair())
                    End If
                ElseIf ifsc IsNot Nothing AndAlso ifsc.IsSectionHeaderVisible() Then
                    kvp = New IniFileKeyValuePair(ifsc, reader.ReadKey(), reader.ReadValue())
                    If reader.CurrentChar = ";"c Then
                        kvp.Comment = reader.ReadComment()
                    End If
                    ifsc.KeyValues.Add(kvp)
                Else
                    Throw New IniFileSyntaxException("Comment or section expected.")
                End If
            Loop While reader.Advance()
        End If
        reader.Flush()
    End Sub

    ''' <summary>
    ''' Saves this ini file into the specified stream.
    ''' </summary>
    ''' <param name="stream">Specifies the stream into which to write the contents of this ini file.</param>
    Public Sub Save(ByVal stream As IO.Stream)
        Dim sw As New IniFileCustomWriter(stream, CommentSeparator, KeyValueSeparator)
        For Each s As IniFileSection In Sections
            sw.WriteSection(s)
        Next
        sw.Flush()
    End Sub

    ''' <summary>
    ''' Saves this ini file into the specified stream encoding the ini file's content as specified.
    ''' </summary>
    ''' <param name="stream">Specifies the stream into which to write the contents of this ini file.</param>
    ''' <param name="encoding">Specifies the encoding of the data stored in the stream.</param>
    Public Sub Save(ByVal stream As IO.Stream, ByVal encoding As Encoding)
        Dim sw As New IniFileCustomWriter(stream, encoding, CommentSeparator, KeyValueSeparator)
        For Each s As IniFileSection In Sections
            sw.WriteSection(s)
        Next
        sw.Flush()
    End Sub

    ''' <summary>
    ''' Saves this ini file into the specified stream.
    ''' </summary>
    ''' <param name="filename">Specifies the file into which to write the contents of this ini file.</param>
    Public Sub Save(ByVal fileName As String)
        Dim stream As IO.Stream = New IO.FileStream(fileName, IO.FileMode.Create)
        Save(stream)
        stream.Close()
    End Sub

    ''' <summary>
    ''' Saves this ini file into the specified stream encoding the ini file's content as specified.
    ''' </summary>
    ''' <param name="fileName">Specifies the file into which to write the contents of this ini file.</param>
    ''' <param name="encoding">Specifies the encoding of the data stored in the stream.</param>
    Public Sub Save(ByVal fileName As String, ByVal encoding As Encoding)
        Dim stream As IO.Stream = New IO.FileStream(fileName, IO.FileMode.Create)
        Save(stream, encoding)
        stream.Close()
    End Sub

    ''' <summary>
    ''' Saves this ini file into the specified file ini text writer.
    ''' </summary>
    ''' <param name="writer">Specifies the writer to use when wrriting the ini file's content.</param>
    Public Sub Save(ByVal writer As IniTextWriter)
        For Each s As IniFileSection In Sections
            writer.WriteSection(s)
        Next
        writer.Flush()
    End Sub

    ''' <summary>
    ''' Loads ini data from the specified stream and keeps previously read data.
    ''' </summary>
    ''' <param name="text">Specifies the String from which to read data.</param>
    Public Sub LoadText(ByVal text As String)
        LoadText(text, True, 8192)
    End Sub

    ''' <summary>
    ''' Loads ini data from the specified stream and clears previously read data if specified.
    ''' </summary>
    ''' <param name="text">Specifies the String from which to read data.</param>
    ''' <param name="keepPrevious">Specifies whether to clear preivously read data.</param>
    Public Sub LoadText(ByVal text As String, ByVal keepPrevious As Boolean)
        LoadText(text, keepPrevious, 8192)
    End Sub

    ''' <summary>
    ''' Loads ini data from the specified stream and clears previously read data if specified. The used string buffer's default size is as specified.
    ''' </summary>
    ''' <param name="text">Specifies the String from which to read data.</param>
    ''' <param name="keepPrevious">Specifies whether to clear preivously read data.</param>
    ''' <param name="stringBufferSize">Specifies the default string buffer size used in the string reader.</param>
    Public Sub LoadText(ByVal text As String, ByVal keepPrevious As Boolean, ByVal stringBufferSize As Integer)
        If Not keepPrevious Then
            Sections.Clear()
        End If
        Load(New IniStringReader(text, stringBufferSize))
    End Sub

    ''' <summary>
    ''' Initializes a new instance of IniLibCore.IniFile type.
    ''' </summary>
    Public Sub New()
        iffFlags = IniFileFlags.Default
        ifscSections = CreateSectionCollection()
    End Sub

    ''' <summary>
    ''' Initializes a new instance of IniLibCore.IniFile type.
    ''' </summary>
    ''' <param name="flags">Specifies rules to base the ini-file management on.</param>
    Public Sub New(ByVal flags As IniFileFlags)
        iffFlags = flags
        ifscSections = CreateSectionCollection()
    End Sub

    ''' <summary>
    ''' Initializes a new instance of IniLibCore.IniFile type.
    ''' </summary>
    ''' <param name="fileName">An ini-file to load on startup.</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal fileName As String)
        iffFlags = IniFileFlags.Default
        ifscSections = CreateSectionCollection()
        LoadFile(fileName)
    End Sub

    ''' <summary>
    ''' Initializes a new instance of IniLibCore.IniFile type.
    ''' </summary>
    ''' <param name="fileName">An ini-file to load on startup.</param>
    ''' <param name="flags">Specifies rules to base the ini-file management on.</param>
    ''' <remarks></remarks>
    Public Sub New(ByVal fileName As String, ByVal flags As IniFileFlags)
        iffFlags = flags
        ifscSections = CreateSectionCollection()
        LoadFile(fileName)
    End Sub

    ''' <summary>
    ''' Gets or creates a section with the specified section name.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name of the section to get or create.</param>
    ''' <returns>Returns the section whos section name equals the specified value.</returns>
    Public Function CreateSection(ByVal sectionName As String) As IniFileSection
        Dim sectionComparer As IEqualityComparer(Of String) = ifscSections.Comparer

        For Each s As IniFileSection In ifscSections
            If sectionComparer.Equals(s.SectionName, sectionName) Then
                Return s
            End If
        Next
        CreateSection = New IniFileSection(Me, sectionName)
        Me.Sections.Add(CreateSection)
    End Function

    ''' <summary>
    ''' Gets or creates the section that is specified by the section name, gets or creates a key-value pair with the specified key and sets the value of the key-value pair to the specified value.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name of the section in which to look for the key-value pair with the specified key.</param>
    ''' <param name="key">Specifies the key of the key-value pair whose value to set.</param>
    ''' <param name="newValue">Specifies the new value of the key-value pair to look for.</param>
    ''' <returns>Returns the key-value pair whose value was set.</returns>
    Public Function SetValue(ByVal sectionName As String, ByVal key As String, ByVal newValue As String) As IniFileKeyValuePair
        Dim section As IniFileSection = Nothing
        Dim sectionComparer As IEqualityComparer(Of String) = ifscSections.Comparer

        For Each s As IniFileSection In ifscSections
            If sectionComparer.Equals(s.SectionName, sectionName) Then
                section = s
                Exit For
            End If
        Next
        If section Is Nothing Then
            section = New IniFileSection(Me, sectionName)
            Sections.Add(section)
        Else
            Dim keyValuePairComparer As IEqualityComparer(Of String) = section.KeyValues.Comparer
            For Each kvp As IniFileKeyValuePair In section.KeyValues
                If keyValuePairComparer.Equals(kvp.Key, key) Then
                    kvp.Value = newValue
                    Return kvp
                End If
            Next
        End If
        SetValue = New IniFileKeyValuePair(key, newValue)
        section.KeyValues.Add(SetValue)
    End Function

    ''' <summary>
    ''' Returns the comment entry at the specified index.
    ''' </summary>
    ''' <param name="index">Specifies the index of the comment entry to find. The index is relative to all comment entries.</param>
    Public Function GetCommentEntryAt(ByVal index As Integer) As IniFileKeyValuePair
        Dim ind As Integer
        For Each nd As IniFileSection In Me.Sections
            For Each kvp As IniFileKeyValuePair In nd.KeyValues
                If kvp.IsComment Then
                    If ind = index Then
                        Return kvp
                    End If
                    ind += 1
                End If
            Next
        Next
        Throw New ArgumentOutOfRangeException("Index")
    End Function

    ''' <summary>
    ''' Finds a key that is contained in a section with the specified name.
    ''' </summary>
    ''' <param name="section">Specifies the section name to look for.</param>
    ''' <param name="key">Specifies the key to look for.</param>
    Public Function GetKey(ByVal section As String, ByVal key As String) As IniFileKeyValuePair
        Return Sections(section).KeyValues(key)
    End Function

    ''' <summary>
    ''' Finds a key-value pair that is contained in a section with the specified name.
    ''' </summary>
    ''' <param name="sectionName">Specifies the section name to look for.</param>
    ''' <param name="key">Specifies the key of the key-value pair to look for.</param>
    ''' <param name="value">Specifies the value of the key-value pair to look for.</param>
    Public Function GetKey(ByVal sectionName As String, ByVal key As String, ByVal value As String) As IniFileKeyValuePair
        Dim section As IniFileSection = Sections(sectionName)
        If section IsNot Nothing Then
            Return section.GetKeyValuePair(key, value)
        Else
            Return Nothing
        End If
    End Function

    ''' <summary>
    ''' Finds all keys that are contained in sections with the specified section name.
    ''' </summary>
    ''' <param name="section">Specifies the section name to look for.</param>
    ''' <param name="key">Specifies the key to look for.</param>
    Public Function GetKeys(ByVal section As String, ByVal key As String) As IniFileKeyValuePair()
        Dim dl As New Generic.LinkedList(Of IniFileKeyValuePair)
        Dim bta() As IniFileKeyValuePair
        Dim sectioncomp As IEqualityComparer(Of String) = ifscSections.Comparer
        Dim keycomp As IEqualityComparer(Of String)

        For Each k As IniFileSection In Me.Sections
            If sectioncomp.Equals(section, k.SectionName) Then
                keycomp = k.KeyValues.Comparer
                For Each kvp As IniFileKeyValuePair In k.KeyValues
                    If keycomp.Equals(kvp.Key, key) Then
                        dl.AddLast(kvp)
                    End If
                Next
            End If
        Next
        bta = New IniFileKeyValuePair(dl.Count - 1) {}
        dl.CopyTo(bta, 0)
        Return bta
    End Function

    ''' <summary>
    ''' Gets a section with the specified section name at the specified index.
    ''' </summary>
    ''' <param name="section">Specifies the section name to look for.</param>
    ''' <param name="index">Specifies the index of the section relative to all sections with the specified name.</param>
    Public Function GetSection(ByVal section As String, ByVal index As Integer) As IniFileSection
        Dim sectioncomp As IEqualityComparer(Of String) = ifscSections.Comparer
        Dim ind As Integer
        For Each sc As IniFileSection In Me.Sections
            If sectioncomp.Equals(section, sc.SectionName) Then
                If ind = index Then
                    Return sc
                End If
                ind += 1
            End If
        Next
        Throw New ArgumentOutOfRangeException("Index")
    End Function

    ''' <summary>
    ''' Finds all sections whose section name equals the specified value.
    ''' </summary>
    ''' <param name="section">Specifies the section name to look for.</param>
    Public Function GetSections(ByVal section As String) As IniFileSection()
        Dim dl As New Generic.LinkedList(Of IniFileSection)()
        Dim bta() As IniFileSection
        Dim sectioncomp As IEqualityComparer(Of String) = ifscSections.Comparer
        For Each k As IniFileSection In Me.Sections
            If sectioncomp.Equals(section, k.SectionName) Then
                dl.AddLast(k)
            End If
        Next
        bta = New IniFileSection(dl.Count - 1) {}
        dl.CopyTo(bta, 0)
        Return bta
    End Function

    ''' <summary>
    ''' Gets the index of the section within all sections whose name is equal to the specified section's name.
    ''' </summary>
    ''' <param name="section">Specifies the section whose index to get.</param>
    Public Function GetSectionWithSectionNameIndex(ByVal section As IniFileSection) As Integer
        Dim sectioncomp As IEqualityComparer(Of String) = ifscSections.Comparer
        Dim name As String = section.SectionName
        Dim ind As Integer
        For Each sc As IniFileSection In Me.Sections
            If sc IsNot section Then
                If sectioncomp.Equals(name, sc.SectionName) Then
                    ind += 1
                End If
            Else
                Return ind
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Gets the index of the first section with the specified section name relatively to all sections.
    ''' </summary>
    ''' <param name="section">Specifies the name of the section to find.</param>
    Public Function GetSectionIndexFromSectionName(ByVal section As String) As Integer
        Dim sectioncomp As IEqualityComparer(Of String) = ifscSections.Comparer
        For i As Integer = 0 To Me.Sections.Count - 1
            If sectioncomp.Equals(section, Me.Sections(i).SectionName) Then
                Return i
            End If
        Next
        Return -1
    End Function

    ''' <summary>
    ''' Returns a string representation of the ini file.
    ''' </summary>
    Public Overrides Function ToString() As String
        Dim sb As New StringBuilder(ifscSections.Count * 256)
        ToString(sb)
        Return sb.ToString()
    End Function

    ''' <summary>
    ''' Writes the ini file's contents to the specified string builder.
    ''' </summary>
    ''' <param name="builder">Specifies the string builder that receives the ini file's contents.</param>
    Public Overridable Overloads Sub ToString(ByVal builder As StringBuilder)
        For Each s As IniFileSection In Me.Sections
            If s.IsSectionHeaderVisible Then
                builder.Append("["c)
                builder.Append(s.SectionName)
                builder.Append("]"c)
            End If
            If s.Comment IsNot Nothing Then
                builder.Append(";"c)
                builder.Append(s.Comment)
            End If
            builder.AppendLine()
            For Each kvp As IniFileKeyValuePair In s.KeyValues
                If kvp.IsComment Then
                    builder.Append(";"c)
                    builder.Append(kvp.Comment)
                ElseIf kvp.IsEmpty = False Then
                    builder.Append(kvp.Key)
                    builder.Append(KeyValueSeparator)
                    builder.Append(kvp.Value)
                    If kvp.Comment IsNot Nothing Then
                        builder.Append(CommentSeparator)
                        builder.Append(kvp.Comment)
                    End If
                End If
                builder.AppendLine()
            Next
        Next
    End Sub

    Private Class IniFileCustomReader
        Inherits IniStreamReader

        Private chrCommentSeparator As Char
        Private chrKeyValueSeparator As Char

#Region "Inherited constructors"
        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal encoding As Text.Encoding, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, encoding)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal encoding As Text.Encoding, ByVal byteBufferSize As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, encoding, byteBufferSize)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal encoding As Text.Encoding, ByVal byteBufferSize As Integer, ByVal stringBufferDefaultLength As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, encoding, byteBufferSize, stringBufferDefaultLength)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal byteBufferSize As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, byteBufferSize)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal byteBufferSize As Integer, ByVal stringBufferDefaultLength As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, byteBufferSize, stringBufferDefaultLength)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub
#End Region

        Public Overrides ReadOnly Property CommentSeparator As Char
            Get
                Return chrCommentSeparator
            End Get
        End Property

        Public Overrides ReadOnly Property KeyValueSeparator As Char
            Get
                Return chrKeyValueSeparator
            End Get
        End Property
    End Class

    Private Class IniFileCustomWriter
        Inherits IniStreamWriter

        Private chrCommentSeparator As Char
        Private chrKeyValueSeparator As Char

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal bufferSize As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, bufferSize)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal encoding As Encoding, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, encoding)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Sub New(ByVal underlyingStream As IO.Stream, ByVal encoding As Encoding, ByVal bufferSize As Integer, ByVal commentSeparator As Char, ByVal keyValueSeparator As Char)
            MyBase.New(underlyingStream, encoding, bufferSize)
            chrCommentSeparator = commentSeparator
            chrKeyValueSeparator = keyValueSeparator
        End Sub

        Public Overrides ReadOnly Property CommentSeparator As Char
            Get
                Return chrCommentSeparator
            End Get
        End Property

        Public Overrides ReadOnly Property KeyValueSeparator As Char
            Get
                Return chrKeyValueSeparator
            End Get
        End Property
    End Class
End Class
