﻿Imports System.Text

''' <summary>
''' Defines an ini file section that contains a section header and key-value pairs.
''' </summary>
''' <remarks></remarks>
<Serializable()> _
Public Class IniFileSection
    ''' <summary>
    ''' Occurs when the ini file section's name changes.
    ''' </summary>
    Public Event SectionNameChanged As EventHandler
    ''' <summary>
    ''' Occurs when the ini file section's comment changes.
    ''' </summary>
    Public Event SectionCommentChanged As EventHandler

    Private strSection As String
    Private ifFile As IniFile
    Private ifkvpcKeyValues As IniFileKeyValuePairCollection
    Private strComment As String

    ''' <summary>
    ''' Raises the SectionNameChanged event.
    ''' </summary>
    ''' <param name="e">An EventArgs that contains the event data.</param>
    Protected Overridable Sub OnSectionNameChanged(ByVal e As EventArgs)
        RaiseEvent SectionNameChanged(Me, e)
    End Sub

    ''' <summary>
    ''' Raises the SectionCommentChanged event.
    ''' </summary>
    ''' <param name="e">An EventArgs that contains the event data.</param>
    Protected Overridable Sub OnSectionCommentChanged(ByVal e As EventArgs)
        RaiseEvent SectionCommentChanged(Me, e)
    End Sub

    ''' <summary>
    ''' A collection of the ini key-value pairs contained in the current ini section.
    ''' </summary>
    Public ReadOnly Property KeyValues As IniFileKeyValuePairCollection
        Get
            Return ifkvpcKeyValues
        End Get
    End Property

    ''' <summary>
    ''' Returns whether the section name is visible. If this value is false the section either is a comment or empty.
    ''' </summary>
    Public ReadOnly Property IsSectionNameVisible As Boolean
        Get
            Return Not (String.IsNullOrEmpty(strSection))
        End Get
    End Property

    ''' <summary>
    ''' Returns whether the section header is visible.
    ''' </summary>
    Public ReadOnly Property IsSectionHeaderVisible As Boolean
        Get
            Return Not (String.IsNullOrEmpty(SectionName) AndAlso strComment Is Nothing)
        End Get
    End Property

    ''' <summary>
    ''' Returns whether the section is a comment section.
    ''' </summary>
    Public ReadOnly Property IsComment As Boolean
        Get
            Return (String.IsNullOrEmpty(strSection) AndAlso strComment IsNot Nothing)
        End Get
    End Property

    ''' <summary>
    ''' Returns whether the section's comment is visible.
    ''' </summary>
    Public ReadOnly Property IsCommentVisible As Boolean
        Get
            Return (strComment IsNot Nothing)
        End Get
    End Property

    ''' <summary>
    ''' Gets or sets the section's name.
    ''' </summary>
    Public Property SectionName As String
        Get
            Return strSection
        End Get
        Set(ByVal value As String)
            If ifFile.IsValidSectionName(value) Then
                If ifFile IsNot Nothing Then
                    If (ifFile.Flags And IniFileFlags.SupportsKeyValuePairKeyRepetition) <> IniFileFlags.SupportsKeyValuePairKeyRepetition Then
                        If ifFile.Sections.Contains(value) Then
                            Throw New ArgumentException("The specified section name is already present in the ini file. Multiple sections that have the same name are not allowed by the ini file.")
                        End If
                    End If
                End If
                strSection = value
                If IsSectionHeaderVisible = False Then
                    For Each f As IniFileKeyValuePair In KeyValues
                        If f.IsCommentOrEmpty = False Then
                            Throw New ArgumentException("Section name can not be invisible as the section contains key-value pairs whose key-value part is visible.")
                        End If
                    Next
                End If
                OnSectionNameChanged(New EventArgs())
            Else
                Throw New IniFileSyntaxException("Invalid section name.")
            End If
        End Set
    End Property

    ''' <summary>
    ''' Gets the amount of entries that are represented by this ini file section. This usually is the amount of lines represented by this ini file section.
    ''' </summary>
    Public Function GetEntryCount() As Integer
        If IsSectionHeaderVisible Then
            Return Me.KeyValues.Count + 1
        Else
            Return Me.KeyValues.Count
        End If
    End Function

    ''' <summary>
    ''' Gets or sets the comment describing the current section.
    ''' </summary>
    Public Property Comment As String
        Get
            Return strComment
        End Get
        Set(ByVal value As String)
            If ifFile.IsValidComment(value) Then
                strComment = value
                OnSectionCommentChanged(New EventArgs())
            Else
                Throw New IniFileSyntaxException("Invalid comment text.")
            End If
        End Set
    End Property

    ''' <summary>
    ''' Gets the ini file containing the current ini section.
    ''' </summary>
    Public ReadOnly Property File As IniFile
        Get
            Return ifFile
        End Get
    End Property

    ''' <summary>
    ''' Gets the first ini key-value pair with the specified key.
    ''' </summary>
    ''' <param name="key">Specifies the key of the ini key-value pair to find.</param>
    Public Function GetKeyValuePair(ByVal key As String) As IniFileKeyValuePair
        Dim cmp As IEqualityComparer(Of String) = KeyValues.Comparer

        For i As Integer = 0 To ifkvpcKeyValues.Count - 1
            If cmp.Equals(ifkvpcKeyValues(i).Key, key) Then
                Return ifkvpcKeyValues(i)
            End If
        Next
        Return Nothing
    End Function

    ''' <summary>
    ''' Gets the first ini key-value pair equal to the specified key-value pair.
    ''' </summary>
    ''' <param name="key">Specifies the key of the key-value pair to find.</param>
    ''' <param name="value">Specifies the value of the key-value pair to find.</param>
    Public Function GetKeyValuePair(ByVal key As String, ByVal value As String) As IniFileKeyValuePair
        Dim cmp As IEqualityComparer(Of String) = KeyValues.Comparer
        For i As Integer = 0 To ifkvpcKeyValues.Count - 1
            If (ifkvpcKeyValues(i) Is value OrElse String.Equals(ifkvpcKeyValues(i).Value, value, StringComparison.InvariantCulture)) AndAlso cmp.Equals(ifkvpcKeyValues(i).Key, key) Then
                Return ifkvpcKeyValues(i)
            End If
        Next
        Return Nothing
    End Function

    ''' <summary>
    ''' Gets the first or creates ini key-value pair with the specified key.
    ''' </summary>
    ''' <param name="key">Specifies the key of the ini key-value pair to find.</param>
    Public Function CreateKeyValuePair(ByVal key As String) As IniFileKeyValuePair
        Dim cmp As IEqualityComparer(Of String) = KeyValues.Comparer

        For i As Integer = 0 To ifkvpcKeyValues.Count - 1
            If cmp.Equals(ifkvpcKeyValues(i).Key, key) Then
                Return ifkvpcKeyValues(i)
            End If
        Next
        CreateKeyValuePair = New IniFileKeyValuePair(Me, key)
        Me.KeyValues.Add(CreateKeyValuePair)
    End Function

    ''' <summary>
    ''' Gets the first or creates ini key-value pair with the specified key and sets its value to the specified value.
    ''' </summary>
    ''' <param name="key">Specifies the key of the ini key-value pair to find.</param>
    ''' <param name="newValue">Specifies the new value of the ini key-value pair.</param>
    Public Function CreateKeyValuePair(ByVal key As String, ByVal newValue As String) As IniFileKeyValuePair
        Dim cmp As IEqualityComparer(Of String) = KeyValues.Comparer

        For i As Integer = 0 To ifkvpcKeyValues.Count - 1
            If cmp.Equals(ifkvpcKeyValues(i).Key, key) Then
                CreateKeyValuePair = ifkvpcKeyValues(i)
                CreateKeyValuePair.Value = newValue
                Return CreateKeyValuePair
            End If
        Next
        CreateKeyValuePair = New IniFileKeyValuePair(Me, key)
        CreateKeyValuePair.Value = newValue
        Me.KeyValues.Add(CreateKeyValuePair)
    End Function

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String)
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name, a describing comment and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    ''' <param name="comment">Specifies a comment describing the current section.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String, ByVal comment As String)
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
        Me.Comment = comment
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name and default key-value pairs and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    ''' <param name="keyValuePairs">Specifies default key-value pairs of the ini section.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String, ByVal keyValuePairs As IEnumerable(Of IniFileKeyValuePair))
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
        For Each kvp As IniFileKeyValuePair In keyValuePairs
            KeyValues.Add(kvp)
        Next
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name, a describing comment, default key-value pairs and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    ''' <param name="comment">Specifies a comment describing the current section.</param>
    ''' <param name="keyValuePairs">Specifies default key-value pairs of the ini section.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String, ByVal comment As String, ByVal keyValuePairs As IEnumerable(Of IniFileKeyValuePair))
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
        Me.Comment = comment
        For Each kvp As IniFileKeyValuePair In keyValuePairs
            KeyValues.Add(kvp)
        Next
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name, a describing comment and default key-value pairs and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    ''' <param name="keyValuePairs">Specifies default key-value pairs of the ini section.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String, ByVal keyValuePairs As IEnumerable(Of IniKeyValuePairDescription))
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
        For Each kvp As IniKeyValuePairDescription In keyValuePairs
            KeyValues.Add(New IniFileKeyValuePair(Me, kvp))
        Next
    End Sub

    ''' <summary>
    ''' Initializes a new instance of the IniLib.IniFileSection class with the specified section name, a describing comment and default key-value pairs and sets the parent file to the specified value.
    ''' </summary>
    ''' <param name="parent">Specifies the parent file containing the current ini section.</param>
    ''' <param name="sectionName">Specifies the section's names.</param>
    ''' <param name="comment">Specifies a comment describing the current section.</param>
    ''' <param name="keyValuePairs">Specifies default key-value pairs of the ini section.</param>
    Public Sub New(ByVal parent As IniFile, ByVal sectionName As String, ByVal comment As String, ByVal keyValuePairs As IEnumerable(Of IniKeyValuePairDescription))
        ifFile = parent
        ifkvpcKeyValues = parent.CreateKeyValuePairCollection(Me)
        Me.SectionName = sectionName
        Me.Comment = comment
        For Each kvp As IniKeyValuePairDescription In keyValuePairs
            KeyValues.Add(New IniFileKeyValuePair(Me, kvp))
        Next
    End Sub

    ''' <summary>
    ''' Convert this ini file section and its key-value pairs into a string representation.
    ''' </summary>
    Public Overrides Function ToString() As String
        Dim sb As New StringBuilder(ifkvpcKeyValues.Count * 32)
        If Not String.IsNullOrEmpty(SectionName) Then
            sb.Append("["c)
            sb.Append(SectionName)
            sb.Append("]"c)
            If Comment IsNot Nothing Then
                If ifFile Is Nothing Then
                    sb.Append(";"c)
                Else
                    sb.Append(ifFile.CommentSeparator)
                End If
                sb.Append(Comment)
            End If
            sb.AppendLine()
        End If
        For Each k As IniFileKeyValuePair In ifkvpcKeyValues
            sb.AppendLine(k.Value)
        Next
        Return sb.ToString()
    End Function
End Class
